#------------------------------------------------------------------------------
#
#   cr6809.py
#
#   Package module file for the SB-Assembler sbasm
#   See www.sbprojects.com for details
#
#   This cross overlay will handle the assembling of 6809
#
#------------------------------------------------------------------------------

import sys, os
import assem, dec, errors, target

crossversion = '3.00.00'
minversion = '3.00.00'

#------------------------------------------------------------------------------

def Help():

    print 'Tell a little story about this cross overlay'

#------------------------------------------------------------------------------

def CrossInit():

    """
    Initialize this cross overlay.
    """

    global Asm, Flags
    
    assem.CheckVersions(crossversion, minversion)

    dec.Asm.Instructions = {
        'ABX'  : (Implied,'3A','3'),
        'ASLA' : (Implied,'48','2'),
        'ASLB' : (Implied,'58','2'),
        'ASRA' : (Implied,'47','2'),
        'ASRB' : (Implied,'57','2'),
        'CLRA' : (Implied,'4F','2'),
        'CLRB' : (Implied,'5F','2'),
        'COMA' : (Implied,'43','2'),
        'COMB' : (Implied,'53','2'),
        'DAA'  : (Implied,'19','2'),
        'DECA' : (Implied,'4A','2'),
        'DECB' : (Implied,'5A','2'),
        'INCA' : (Implied,'4C','2'),
        'INCB' : (Implied,'5C','2'),
        'LSLA' : (Implied,'48','2'),
        'LSLB' : (Implied,'58','2'),
        'LSRA' : (Implied,'44','2'),
        'LSRB' : (Implied,'54','2'),
        'MUL'  : (Implied,'3D','11'),
        'NEGA' : (Implied,'40','2'),
        'NEGB' : (Implied,'50','2'),
        'NOP'  : (Implied,'12','2'),
        'ROLA' : (Implied,'49','2'),
        'ROLB' : (Implied,'59','2'),
        'RORA' : (Implied,'46','2'),
        'RORB' : (Implied,'56','2'),
        'RTI'  : (Implied,'3B','15'),
        'RTS'  : (Implied,'39','5'),
        'SEX'  : (Implied,'1D','2'),
        'SWI'  : (Implied,'3F','19'),
        'SWI2' : (Implied,'103F','20'),
        'SWI3' : (Implied,'113F','20'),
        'SYNC' : (Implied,'13','2'),
        'TSTA' : (Implied,'4D','2'),
        'TSTB' : (Implied,'5D','2'),

        'ABA'  : (Implied,'3404ABE0','12'),
        'ASLD' : (Implied,'5849','4'),
        'ASRD' : (Implied,'4756','4'),
        'CBA'  : (Implied,'3404A1E0','12'),
        'CLC'  : (Implied,'1CFE','3'),
        'CLF'  : (Implied,'1CBF','3'),
        'CLI'  : (Implied,'1CEF','3'),
        'CLIF' : (Implied,'1CAF','3'),
        'CLRD' : (Implied,'4F5F','4'),
        'CLV'  : (Implied,'1CFD','3'),
        'COMD' : (Implied,'4353','4'),
        'DES'  : (Implied,'327F','5'),
        'DEX'  : (Implied,'301F','5'),
        'DEY'  : (Implied,'313F','5'),
        'DEU'  : (Implied,'335F','5'),
        'INS'  : (Implied,'3261','5'),
        'INU'  : (Implied,'3341','5'),
        'INX'  : (Implied,'3001','5'),
        'INY'  : (Implied,'3121','5'),
        'LSLD' : (Implied,'5849','4'),
        'LSRD' : (Implied,'4456','4'),
        'PSHA' : (Implied,'3402','6'),
        'PSHB' : (Implied,'3404','6'),
        'PSHCC': (Implied,'3401','6'),
        'PSHD' : (Implied,'3406','7'),
        'PSHDP': (Implied,'3408','6'),
        'PSHPC': (Implied,'3480','6'),
        'PSHX' : (Implied,'3410','6'),
        'PSHY' : (Implied,'3420','6'),
        'PULA' : (Implied,'3502','6'),
        'PULB' : (Implied,'3504','6'),
        'PULCC': (Implied,'3501','6'),
        'PULD' : (Implied,'3506','7'),
        'PULDP': (Implied,'3508','6'),
        'PULPC': (Implied,'3580','6'),
        'PULX' : (Implied,'3510','6'),
        'PULY' : (Implied,'3520','6'),
        'ROLD' : (Implied,'5949','4'),
        'RORD' : (Implied,'4656','4'),
        'SBA'  : (Implied,'3404A0E0','12'),
        'SEC'  : (Implied,'1A01','3'),
        'SEF'  : (Implied,'1A40','3'),
        'SEI'  : (Implied,'1A10','3'),
        'SEIF' : (Implied,'1A50','3'),
        'SEV'  : (Implied,'1A02','3'),
        'TAB'  : (Implied,'1F894D','9'),
        'TAP'  : (Implied,'1F8A','7'),
        'TBA'  : (Implied,'1F984D','9'),
        'TPA'  : (Implied,'1FA8','7'),
        'TSX'  : (Implied,'1F41','7'),
        'TXS'  : (Implied,'1F14','7'),
        'WAI'  : (Implied,'3CFF','20'),

        'BCC'  : (Branch,'24','3'),
        'BCS'  : (Branch,'25','3'),
        'BEQ'  : (Branch,'27','3'),
        'BGE'  : (Branch,'2C','3'),
        'BGT'  : (Branch,'2E','3'),
        'BHI'  : (Branch,'22','3'),
        'BHS'  : (Branch,'24','3'),
        'BLE'  : (Branch,'2F','3'),
        'BLO'  : (Branch,'25','3'),
        'BLS'  : (Branch,'23','3'),
        'BLT'  : (Branch,'2D','3'),
        'BMI'  : (Branch,'2B','3'),
        'BNE'  : (Branch,'26','3'),
        'BPL'  : (Branch,'2A','3'),
        'BRA'  : (Branch,'20','3'),
        'BRN'  : (Branch,'21','3'),
        'BSR'  : (Branch,'8D','7'),
        'BVC'  : (Branch,'28','3'),
        'BVS'  : (Branch,'29','3'),

        'LBCC' : (LBranch,'1024','5+'),
        'LBCS' : (LBranch,'1025','5+'),
        'LBEQ' : (LBranch,'1027','5+'),
        'LBGE' : (LBranch,'102C','5+'),
        'LBGT' : (LBranch,'102E','5+'),
        'LBHI' : (LBranch,'1022','5+'),
        'LBHS' : (LBranch,'1024','5+'),
        'LBLE' : (LBranch,'102F','5+'),
        'LBLO' : (LBranch,'1025','5+'),
        'LBLS' : (LBranch,'1023','5+'),
        'LBLT' : (LBranch,'102D','5+'),
        'LBMI' : (LBranch,'102B','5+'),
        'LBNE' : (LBranch,'1026','5+'),
        'LBPL' : (LBranch,'102A','5+'),
        'LBRA' : (LBranch,'16','55'),
        'LBRN' : (LBranch,'1021','5+'),
        'LBSR' : (LBranch,'17','9'),
        'LBVC' : (LBranch,'1028','5+'),
        'LBVS' : (LBranch,'1029','5+'),

        'ADCA' : (Multi, ('89','99','A9','B9'),('2','4','4','5'),False),
        'ADCB' : (Multi, ('C9','D9','E9','F9'),('2','4','4','5'),False),
        'ADDA' : (Multi, ('8B','9B','AB','BB'),('2','4','4','5'),False),
        'ADDB' : (Multi, ('CB','DB','EB','FB'),('2','4','4','5'),False),
        'ADDD' : (Multi, ('C3','D3','E3','F3'),('4','6','6','7'),True),
        'ANDA' : (Multi, ('84','94','A4','B4'),('2','4','4','5'),False),
        'ANDB' : (Multi, ('C4','D4','E4','F4'),('2','4','4','5'),False),
        'ANDCC' : (Multi, ('1C','','',''),('3','0','0','0'),False),
        'ASL'  : (Multi, ('','08','68','78'),('0','6','6','7'),False),
        'ASR'  : (Multi, ('','07','67','77'),('0','6','6','7'),False),
        'BITA' : (Multi, ('85','95','A5','B5'),('2','4','4','5'),False),
        'BITB' : (Multi, ('C5','D5','E5','F5'),('2','4','4','5'),False),
        'CLR'  : (Multi, ('','0F','6F','7F'),('0','6','6','7'),False),
        'CMPA' : (Multi, ('81','91','A1','B1'),('2','4','4','5'),False),
        'CMPB' : (Multi, ('C1','D1','E1','F1'),('2','4','4','5'),False),
        'CMPD' : (Multi, ('1083','1093','10A3','10B3'),('5','7','7','8'),True),
        'CMPS' : (Multi, ('118C','119C','11AC','11BC'),('5','7','7','8'),True),
        'CMPU' : (Multi, ('1183','1193','11A3','11B3'),('5','7','7','8'),True),
        'CMPX' : (Multi, ('8C','9C','AC','BC'),('4','6','6','7'),True),
        'CMPY' : (Multi, ('108C','109C','10AC','10BC'),('5','7','7','8'),True),
        'COM'  : (Multi, ('','03','63','73'),('0','6','6','7'),False),
        'CWAI' : (Multi, ('3C','','',''),('20','0','0','0'),False),
        'DEC'  : (Multi, ('','0A','6A','7A'),('0','6','6','7'),False),
        'EORA' : (Multi, ('88','98','A8','B8'),('2','4','4','5'),False),
        'EORB' : (Multi, ('C8','D8','E8','F8'),('2','4','4','5'),False),
        'INC'  : (Multi, ('','0C','6C','7C'),('0','6','6','7'),False),
        'JMP'  : (Multi, ('','0E','6E','7E'),('0','3','3','4'),False),
        'JSR'  : (Multi, ('','9D','AD','BD'),('0','7','7','8'),False),
        'LDA'  : (Multi, ('86','96','A6','B6'),('2','4','4','5'),False),
        'LDAA' : (Multi, ('86','96','A6','B6'),('2','4','4','5'),False),
        'LDB'  : (Multi, ('C6','D6','E6','F6'),('2','4','4','5'),False),
        'LDAB' : (Multi, ('C6','D6','E6','F6'),('2','4','4','5'),False),
        'LDD'  : (Multi, ('CC','DC','EC','FC'),('3','5','5','6'),True),
        'LDAD' : (Multi, ('CC','DC','EC','FC'),('3','5','5','6'),True),
        'LDS'  : (Multi, ('10CE','10DE','10EE','10FE'),('4','6','6','7'),True),
        'LDU'  : (Multi, ('CE','DE','EE','FE'),('3','5','5','6'),True),
        'LDX'  : (Multi, ('8E','9E','AE','BE'),('3','5','5','6'),True),
        'LDY'  : (Multi, ('108E','109E','10AE','10BE'),('4','6','6','7'),True),
        'LEAS' : (Multi, ('','','32',''),('0','0','4','0'),False),
        'LEAU' : (Multi, ('','','33',''),('0','0','4','0'),False),
        'LEAX' : (Multi, ('','','30',''),('0','0','4','0'),False),
        'LEAY' : (Multi, ('','','31',''),('0','0','4','0'),False),
        'LSL'  : (Multi, ('','08','68','78'),('0','6','6','7'),False),
        'LSR'  : (Multi, ('','04','64','74'),('0','6','6','7'),False),
        'NEG'  : (Multi, ('','00','60','70'),('0','6','6','7'),False),
        'ORA'  : (Multi, ('8A','9A','AA','BA'),('2','4','4','5'),False),
        'ORB'  : (Multi, ('CA','DA','EA','FA'),('2','4','4','5'),False),
        'ORCC' : (Multi, ('1A','','',''),('3','0','0','0'),False),
        'ROL'  : (Multi, ('','09','69','79'),('0','6','6','7'),False),
        'ROR'  : (Multi, ('','06','66','76'),('0','6','6','7'),False),
        'SBCA' : (Multi, ('82','92','A2','B2'),('2','4','4','5'),False),
        'SBCB' : (Multi, ('C2','D2','E2','F2'),('2','4','4','5'),False),
        'STA'  : (Multi, ('','97','A7','B7'),('0','4','4','5'),False),
        'STAA' : (Multi, ('','97','A7','B7'),('0','4','4','5'),False),
        'STB'  : (Multi, ('','D7','E7','F7'),('0','4','4','5'),False),
        'STAB' : (Multi, ('','D7','E7','F7'),('0','4','4','5'),False),
        'STD'  : (Multi, ('','DD','ED','FD'),('0','5','5','6'),False),
        'STAD' : (Multi, ('','DD','ED','FD'),('0','5','5','6'),False),
        'STS'  : (Multi, ('','10DF','10EF','10FF'),('0','6','6','7'),False),
        'STU'  : (Multi, ('','DF','EF','FF'),('0','5','5','6'),False),
        'STX'  : (Multi, ('','9F','AF','BF'),('0','5','5','6'),False),
        'STY'  : (Multi, ('','109F','10AF','10BF'),('0','6','6','7'),False),
        'SUBA' : (Multi, ('80','90','A0','B0'),('2','4','4','5'),False),
        'SUBB' : (Multi, ('C0','D0','E0','F0'),('2','4','4','5'),False),
        'SUBD' : (Multi, ('83','93','A3','B3'),('4','6','6','7'),True),
        'TST'  : (Multi, ('','0D','6D','7D'),('0','6','6','7'),False),

        'PSHS' : (Stack, '34', 5),
        'PSHU' : (Stack, '36', 5),
        'PULS' : (Stack, '35', 5),
        'PULU' : (Stack, '37', 5),        
        
        'EXG'  : (Exchange, '1E','7'),
        'TFR'  : (Exchange, '1F','7')
    }

    errors.Error_List['6809double'] = 'Repeated register(s) ignored'
    
    dec.Asm.Timing_Length = 2

    dec.Asm.Memory = 0
    if dec.Asm.Pass == 1:
        sys.stdout.write('Loaded ' + dec.Cross.Name[2:] + ' overlay version ' + crossversion + dec.EOL)

    dec.Asm.Max_Address = dec.MAX16
    dec.Asm.PP_TA_Factor = 1
    dec.Flags.BigEndian = True
    dec.Asm.DP6809 = 0

    return
    
#------------------------------------------------------------------------------

def CrossDirective():

    """
    One extra directive
    """

    if len(dec.Asm.Mnemonic) > 1:
        directive = dec.Asm.Mnemonic[1:3].upper()
    else:
        directive = dec.Asm.Mnemonic
    
    if directive == 'DP':
        DirDP()
        return True

    return False     # return True if we handled the directive

#------------------------------------------------------------------------------

def CrossCleanUp():

    """
    No need to do any cleaning.
    """

    return    
    
#------------------------------------------------------------------------------

def CrossMnemonic():

    """
    Find and decode the current mnemonic.
    """

    global Asm

    if dec.Asm.Mnemonic in dec.Asm.Instructions:
        func = dec.Asm.Instructions[dec.Asm.Mnemonic][0]
        func()
    else:
        errors.DoError('badopco', False)

#-----------------------------------------------------------------------------

def MissingOperand():

    """
    An operand must follow. Raise an error if it's not.
    """
    
    if dec.Asm.Parse_Pointer == 0:
        errors.DoError('missoper', False)
        return True
    else:
        return False

#-----------------------------------------------------------------------------


def NoMore():

    """
    No more operands shoudl follow. Raise a warning if more parameters follow.
    """

    if assem.MoreParameters():
        errors.DoWarning('extrign', False)
        
#-----------------------------------------------------------------------------

def Implied():

    """
    Handle implied addressing mode.
    Simply save the opcode and we're done with it.
    """

    global Asm

    opcode = int(dec.Asm.Instructions[dec.Asm.Mnemonic][1],16)
    length = len(dec.Asm.Instructions[dec.Asm.Mnemonic][1])
    if length == 8:
        target.CodeByte(opcode >> 24)
        target.CodeByte(opcode >> 16)
        target.CodeByte(opcode >> 8)
    elif length == 6:
        target.CodeByte(opcode >> 16)
        target.CodeByte(opcode >> 8)
    elif length == 4:
        target.CodeByte(opcode >> 8)
    target.CodeByte(opcode)
    dec.Asm.Timing = dec.Asm.Instructions[dec.Asm.Mnemonic][2]

#-----------------------------------------------------------------------------

def Branch():

    """
    Handle branch instructions.
    Displacement is destinaiton - current address - 2
    """

    global Asm

    if MissingOperand():
        return

    value = assem.EvalExpr()
    offset = value[0] - dec.Asm.BOL_Address - 2
    
    if dec.Asm.Pass == 2 and (offset < -128 or offset > 127):
        errors.DoError('range', False)

    target.CodeByte(int(dec.Asm.Instructions[dec.Asm.Mnemonic][1],16))
    target.CodeByte(offset)
    dec.Asm.Timing = dec.Asm.Instructions[dec.Asm.Mnemonic][2]
    
    NoMore()

#-----------------------------------------------------------------------------

def LBranch():

    """
    Handle long branch instructions.
    Displacement is destinaiton - current address - 3 in case 3 byte long instruction
    Displacement is destinaiton - current address - 4 in case 4 byte long instruction
    """

    global Asm

    if MissingOperand():
        return

    opcode = int(dec.Asm.Instructions[dec.Asm.Mnemonic][1],16)
    value = assem.EvalExpr()
    offset = value[0] - dec.Asm.BOL_Address - 3

    if opcode > 255:
        offset = offset - 1
        target.CodeByte(opcode >> 8)
    target.CodeByte(opcode)
    target.CodeWord(offset)
    dec.Asm.Timing = dec.Asm.Instructions[dec.Asm.Mnemonic][2]
    
    NoMore()

#-----------------------------------------------------------------------------

def Multi():

    """
    Handle multiple operand instructions
    This is quite a complex one.
    - First let's see if an immediate value follows the mnemonic
    - If it is not an immediate value, it may be an indirect mode of some sort
    - Next we check if it's r,index mode
    - Then it could be zero offset,index
    - or forced or automatic offset,index mode
    - index,PCR or index,PC mode
    """

    global Asm
    
    if MissingOperand():
        return

    if assem.NowChar() in '#/=\\':
        # It is direct addressing mode
        prefix = assem.NowChar(True)
        opcode = dec.Asm.Instructions[dec.Asm.Mnemonic][1][0]
        if opcode == '':
            errors.DoError('badoper', False)
            return
        
        value = assem.EvalExpr()[0]
        if prefix == '/':
            value = value >> 8
        elif prefix == '=':
            value = value >> 16
        elif prefix == '\\':
            value = value >> 24
        
        Code(int(opcode,16))
        if dec.Asm.Instructions[dec.Asm.Mnemonic][3]:
            target.CodeWord(value)
        else:
            target.CodeByte(value)
        dec.Asm.Timing = dec.Asm.Instructions[dec.Asm.Mnemonic][2][0]
        
        NoMore()
        return

    postbyte = -1    # Indicate non indexed or indirect mode per default
    oplength = 0
    extra = 0
    value = (0,False)

    indexregs  = {'X':0, 'Y':1, 'U':2, 'US':2, 'S':3, 'SP':3}
    incdecregs = {'X+':'80', 'Y+':'A0', 'U+':'C0', 'US+':'C0', 'S+':'E0', 'SP+':'E0', 'X++':'81', 'Y++':'A1', 'U++':'C1', 'US++':'C1', 'S++':'E1', 'SP++':'E1', 
                  '-X':'82', '-Y':'A2', '-U':'C2', '-US':'C2', '-S':'E2', '-SP':'E2', '--X':'83', '--Y':'A3', '--U':'C3', '--US':'C3', '--S':'E3', '--SP':'E3'}

    indirect = assem.NowChar()
    if indirect in ('[','('):
        dec.Asm.Parse_Pointer = dec.Asm.Parse_Pointer + 1
    else:
        indirect = ''

    pointer = dec.Asm.Parse_Pointer
    register = assem.GetWord().upper()
    
    if register in ('A','B','D'):
        # It is r,index mode
        if register == 'A':
            postbyte = int('86',16)
            extra = 1
        elif register == 'B':
            postbyte = int('85',16)
            extra = 1
        else:
            postbyte = int('8B',16)
            extra = 4

        if not assem.MoreParameters():
            errors.DoError('missoper', False)
            return
        
        register = assem.GetWord('','',' ,)]').upper()
        
        if register in indexregs:
            postbyte = postbyte + (indexregs[register] << 5)
        else:
            errors.DoError('badoper',False)
            return

        if indirect != '':
            if assem.NowChar(True).upper() in (')',']'):
                postbyte = postbyte + 16
                extra = extra + 3
            else:
                errors.DoError('badoper',False)
                return
        
    else:
        # It was not r,index mode
        dec.Asm.Parse_Pointer = pointer
    
        prefix = ''
        
        if assem.NowChar() == ',':
            # Zero offset indexed mode
            value = (0,False)
    
        elif assem.NowChar() in '<>':
            # Forced short or long addressing mode
            prefix = assem.NowChar(True)
            value = assem.EvalExpr()
        else:
            value = assem.EvalExpr()

        if assem.NowChar() in (')',']'):
            # Normal indirect mode [address]
            # Ignoring prefix < or >
            dec.Asm.Parse_Pointer = dec.Asm.Parse_Pointer + 1
            postbyte = int('9F',16)
            extra = 5
            oplength = 2
        
        else:
            # Indirect/indexed mode [offset,ix]
            if assem.NowChar() != ' ':
                # Some kind of indirect or indexed mode

                if assem.NowChar(True) != ',':
                    errors.DoError('badoper',False)
                    return
                    
                register = assem.GetWord('','',')] ,').upper()

                if register in incdecregs:
                    # Auto increment or decrement indexed
                    postbyte = int(incdecregs[register],16)

                    if value[0] != 0:
                        errors.DoError('range',False)
                        return
                    
                    if (postbyte & 1) == 0:
                        # single inc or dec
                        extra = 2
                        if indirect != '' or assem.NowChar() in (')',']'):
                            # Indirect mode not allowed for single inc or dec
                            errors.DoError('badoper',False)
                            return
                    else:
                        # double inc or dec
                        extra = 3
                        if indirect != '':
                            # It's indirect mode
                            if assem.NowChar(True) in (')',']'):
                                postbyte = postbyte + 16
                                extra = extra + 3
                            else:
                                errors.DoError('badoper',False)
                                return

                else:
                    # Now it's some kind of offset,index mode (maybe together with indirect)
                    if register in indexregs:
                        # It's offset,r
                        osize = OffsetSize(value, prefix, indirect)
                        if osize == 16:
                            postbyte = int('89',16)
                            oplength = 2
                            extra = 4
                        elif osize == 0:
                            postbyte = int('84',16)
                            oplength = 0
                            extra = 0
                        elif osize == 8:
                            postbyte = int('88',16)
                            oplength = 1
                            extra = 1
                        else:
                            # Offset is now 5 bits long
                            postbyte = value[0] & 31
                            oplength = 0
                            extra = 1
                        postbyte = postbyte + (indexregs[register] << 5)

                    elif register == 'PCR':
                        # It's offset,pc or offset,pcr
                        if register == 'PCR':
                            # It's offset,pcr
                            
                            # Treat the offset as a relative address
                            # First calculate PC just behind the current instruction                        
                            if len(dec.Asm.Instructions[dec.Asm.Mnemonic][1][2]) == 2:
                                # Instruction requires no pre-byte
                                nowpc = dec.Asm.BOL_Address + 3
                            else:
                                # Instruction requires pre-byte
                                nowpc = dec.Asm.BOL_Address + 4
                                
                            rel = value[0] - nowpc
                            short = False
                            
                            if value[1] or prefix == '>':
                                # We don't know the value now or we want 16-bit offsets
                                if prefix == '<':
                                    short = True
                            else:
                                # We could have an 8-bit offset
                                if rel >= -128 and rel <= +127:
                                    # It is indeed a short address
                                    short = True
                                    
                            if not short:
                                # Long instructions are 1 byte longer
                                rel = rel -1
                        
                            value = (rel,value[1])
                            
                            if short:
                                postbyte = int('8C',16)
                                oplength = 1
                                extra = 1
                            else:
                                postbyte = int('8D',16)
                                oplength = 2
                                extra = 5
                        
                    elif register == 'PC':
                        # It's offset,pc or offset,pcr
                        if OffsetSize(value, prefix, indirect) == 16:
                            postbyte = int('8D',16)
                            oplength = 2
                            extra = 5
                        else:
                            # 5-bit and 0-bit offsets don't exist for offset,pc
                            postbyte = int('8C',16)
                            oplength = 1
                            extra = 1
                    
                    else:
                        errors.DoError('badoper',False)
                        return
                        
                    if indirect != '':
                        # It's indirect mode as well
                        postbyte = postbyte + 16
                        extra = extra + 3
                        if not assem.NowChar(True) in (')',']'):
                            errors.DoError('badoper',False)
                            
    if postbyte < 0:
        # Normal direct or extended addressing mode
        if IsDirect(value,prefix):
            opcode = dec.Asm.Instructions[dec.Asm.Mnemonic][1][1]
            if opcode == '':
                errors.DoError('badoper',False)
                return
            Code(int(opcode,16))
            target.CodeByte(value[0])
            dec.Asm.Timing = dec.Asm.Instructions[dec.Asm.Mnemonic][2][1]
        else:
            opcode = dec.Asm.Instructions[dec.Asm.Mnemonic][1][3]
            if opcode == '':
                errors.DoError('badoper',False)
                return
            Code(int(opcode,16))
            target.CodeWord(value[0])
            dec.Asm.Timing = dec.Asm.Instructions[dec.Asm.Mnemonic][2][3]
    else:
        # It was some kind of indirect or indexed mode
        opcode = dec.Asm.Instructions[dec.Asm.Mnemonic][1][2]
        if opcode == '':
            errors.DoError('badoper',False)
            return
        Code(int(opcode,16))
        target.CodeByte(postbyte)
        if oplength == 1:
            target.CodeByte(value[0])
        elif oplength == 2:
            target.CodeWord(value[0])
            
        dec.Asm.Timing = str(int(dec.Asm.Instructions[dec.Asm.Mnemonic][2][3]) + extra)
        
    NoMore()

#-----------------------------------------------------------------------------

def OffsetSize(offset, prefix, indirect):

    if prefix == '>':
        # Forced 16-bit
        return 16

    if prefix == '<':
        # Forced 8-bit?
        if dec.Asm.Pass ==2 and (offset[0] < -128 or offset[0] > +127):
            errors.DoError('range',False)
        return 8

    if offset[1]:
        # Forward referenced label used
        return 16
    
    if offset[0] == 0:
        return 0
        
    if offset[0] < -128 or offset[0] > +127:
        return 16
   
    if indirect != '':
        return 8
        
    if offset[0] < -16 or offset[0] > +15:
        return 8
    else:
        return 5

#-----------------------------------------------------------------------------

def IsDirect(value,prefix):

    if prefix == '>' or value[1]:
        # Forced extended mode or forward referenced label
        direct = False
    elif prefix == '<':
        # Forced direct mode
        direct = True
    elif (value[0] >> 8) == dec.Asm.DP6809:
        # Address is within current direct page
        direct = True
    else:
        # Sorry, it's extended mode
        direct = False
    
    return direct    

#-----------------------------------------------------------------------------

def Stack():

    """
    Handle Push and Pull instructions.
    Any number from of registers and register pairs may be pushed and pulled
    by one instruction, all separated by a comma.
    A warning is issued if you specify a particular register more than once.
    """

    global Asm

    if MissingOperand():
        return
        
    registers = {'CC':1, 'CCR':1, 'A':2, 'B':4, 'D':6, 'A:B':6, 'DP':8, 'DPR':8, 'X':16, 'Y':32, 'U':64, 'US':64, 'S':64, 'SP':64, 'PC':128}

    stack = dec.Asm.Mnemonic[-1]

    postbyte = 0
    doubled = False
    
    while True:
        reg = assem.GetWord().upper()
        if reg in registers:
            if reg == stack:
                # S may not be used in PSHS and PULS
                # U may not be used in PSHU and PULU
                errors.DoError('badoper',False)
                return
            bits = registers[reg]
            if (postbyte & bits) != 0:
                doubled = True
            postbyte = postbyte | bits
        else:
            errors.DoError('badoper',False)
            return

        if not assem.MoreParameters():
            break
    
    if doubled:
        errors.DoWarning('6809double',False)

    count = 0
    for i in range(0,8):
        if (postbyte & (1 << i)) != 0:
            if i > 3:
                count = count + 2
            else:
                count = count + 1
    
    target.CodeByte(int(dec.Asm.Instructions[dec.Asm.Mnemonic][1],16))
    target.CodeByte(postbyte)
    dec.Asm.Timing = str(dec.Asm.Instructions[dec.Asm.Mnemonic][2] + count)

#-----------------------------------------------------------------------------

def Exchange():

    """
    Handle EXG and TFR directives.
    Both directives take 2 operands of either 8-bit size or 16-bit size.
    Sizes of the registers may not be mixed.
    """

    global Asm

    if MissingOperand():
        return

    reg1 = assem.GetWord().upper()

    if not assem.MoreParameters():
        errors.DoError('missoper', False)
        return
    
    reg2 = assem.GetWord().upper()

    words = {'A:B':0, 'D':0, 'X':1, 'Y':2, 'US':3, 'U':3, 'SP':4, 'S':4, 'PC':5}
    bytes = {'A':8, 'B':9, 'CC':10, 'CCR':10, 'DP':11, 'DPR':11}    
    
    if reg1 in words:
        if reg2 in words:
            target.CodeByte(int(dec.Asm.Instructions[dec.Asm.Mnemonic][1],16))
            target.CodeByte((words[reg1] << 4) + words[reg2])
            dec.Asm.Timing = dec.Asm.Instructions[dec.Asm.Mnemonic][2]
        else:
            errors.DoError('badoper',False)
    elif reg1 in bytes:
        if reg2 in bytes:
            target.CodeByte(int(dec.Asm.Instructions[dec.Asm.Mnemonic][1],16))
            target.CodeByte((bytes[reg1] << 4) + bytes[reg2])
            dec.Asm.Timing = dec.Asm.Instructions[dec.Asm.Mnemonic][2]
        else:
            errors.DoError('badoper',False)
    else:
        errors.DoError('badoper',False)
         
    NoMore()
    
#-----------------------------------------------------------------------------

def Code(opcode):

    """
    Write an opcode to the target file.
    If the opcode > 255 then a pre-byte must be saved first
    """
    
    if opcode > 255:
        target.CodeByte(opcode >> 8)
    target.CodeByte(opcode)

#-----------------------------------------------------------------------------

def DirDP():

    """
    Handle the .DP directive, which tells the assembler what direct page is
    to be used.
    No forward referenced labels are allowed.
    """

    global Asm

    if MissingOperand():
        return

    value = assem.EvalExpr()
    
    if value[1]:
        errors.DoError('forwref',False)
    else:
        if value[0] < 0 or value[0] > 255:
            errors.DoError('range',False)
        dec.Asm.DP6809 = value[0]
    
    NoMore()    

#-----------------------------------------------------------------------------

if __name__ == '__main__':
    print
    print "This is a python module, it's not a program."
    print "This module is part of the sbasm package."
    print "Run sbasm instead."
    print
    
